################################################################################
#
# File:         bibutils.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibutils.py,v 1.7 2007/08/06 20:06:21 paepcke Exp $
# Description:  Miscellaneous Utilities
# Author:       Andreas Paepcke, Stanford University
# Created:      Wed Sep 27 10:26:25 1995
# Modified:     Sun Aug 05 15:57:42 2007 (Andreas Paepcke) paepcke@EMIL
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Tue Jul 31 08:18:34 2007 (Andreas Paepcke) paepcke@EMIL
#  Modified getLinkTags() so that if no links are defined,
#  we check whether the entry has a URL field instead. If
#  so, we pretend that we did have a link field with that
#  URL anchored the title.
# Sun Jul 29 19:48:18 2007 (Andreas Paepcke) paepcke@EMIL
#  Added function findEmbeddedURL
# Mon Jul 23 17:47:52 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Modified report() to take an option IO stream parameter.
# Fri Jul 20 11:32:27 2007 (Andreas Paepcke) paepcke@Shark.Stanford.EDU
#  Migrated from deprecated regex to re.
# Mon Mar 31 18:44:58 1997 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Changed error msg "Links should be of the form <...>:<...>:URL..." to not have
#  the '<', '>' chars because they get eaten by html browsers.
# Tue Mar 18 12:27:43 1997 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Added verifyPassword and setPassword.
# Tue Sep 10 17:49:02 1996 (Andreas Paepcke) paepcke@Walrus
#  Added joinSentences
# Thu Jul 18 12:34:31 1996 (Andreas Paepcke) paepcke@Walrus
#  Fixed the permission handling of StrOrFileFd when strings are passed.
# Sun Jul  7 12:16:33 1996 (Andreas Paepcke) paepcke@Stingray
#  added function sortDictBibEntriesByLastName
# Fri Jan 26 08:59:50 1996 (Andreas Paepcke) paepcke@Stingray
#  Added fileConsumptionFeedbackGadget class
# Mon Dec 18 10:28:30 1995 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Added check that link specs in parseLinkSpec have a proper target env. spec.
# Wed Dec  6 22:45:42 1995 (Andreas Paepcke) paepcke@Stingray
#  Fixed getAuthorLastNameFirst to properly handle 'J. Ray Scott'
# Tue Dec  5 08:53:42 1995 (Andreas Paepcke) paepcke@Stingray
#  In keyGenerator: added optional 'prefix' argument.
# Sun Oct 29 15:52:32 1995 (Andreas Paepcke) paepcke@Stingray
#  Added digitsOnly
################################################################################


import bibimport
import bibdb
import sys
import string
import types
import time
import re
import os

# Public functions:

# - report
# - extractLinks
# - parseLinkSpec
# - countMatches
# - getAuthorLastNameFirst
# - getAuthorLastNameLast
# - sortByLastName
# - sortDictBibEntriesByLastName
# - getEntryFldSafely
# - getLinkTags
# - isInstanceOf
# - allSubclassesOf
# - entryNameComparison
# - digitsOnly
# - readNum
# - joinSentences
# - verifyPassword
# - setPassword

# Public classes and methods:

# KeyGenerator:
#   generateKey

# feedbackGadget/fileConsumptionFeedbackGadget:
#   tick
#   reset

# StrOrFileFD
#   open
#   close
#   flush
#   isatty
#   read
#   readline
#   readlines
#   seek
#   tell
#   write
#   writelines
#   isFile
#   len


#####################################################
#
# Exceptions
#
#####################################################

class InterbibException(Exception):
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)
  def message(self):
    return str(self.value)

class BadLinkSpecFormat(InterbibException):
  pass

class FaultyPasswordMechanism(InterbibException):
  pass

#####################################################
#
# Private Constants
#
#####################################################

#                                              |   |   
#                                              v   v
# Separator in links of bib files. As in 'title:www:http://db'
linkComponentSeparator = ':'

prevKeyTime = 0
keyUniquifier = 0

# These following ones are actually defined in bibconvert.py, but
# we have a cross-import problem between this
# bibutils.py file and bibconvert.py. If you change
# the constants here, change them as well in bibconvert.py!!

linksFieldName = 'links'
urlFieldName = 'url'

# Regular expression to find a URL buried in a string:
#   
#
#      Legal URL chars are:
# 	  [0-9a-zA-Z]
# 	  $-_.+!*'()&/:;=?@
# 	  also: Comma (",")
#
#      Legal URL schemes are:
#        http, https, ftp, file, mailto, news, nntp, telnet
#
#      Python regex metachars that must be escaped:
#        ^ $ \ / ( ) | ? + * [ ] { } , . 

urlFinderREPattern = "(http|https|ftp|file|mailto|news|nntp|telnet):\/\/[0-9a-zA-Z\$-_\.\+!\*'\(\)\,&\/:;=\?@]*"
urlFinder = re.compile(urlFinderREPattern)

#####################################################
#
# Utility Functions
#
#####################################################

# Join two sentences together and return one sentence:

def joinSentences(str1, str2):
  str1 = string.strip(str1)
  str2 = string.strip(str2)
  if not str1:
    return str2
  if not str2:
    return str1
  if str1[-1] not in '.,;!?:':
    str1 = str1 + '.'
  return(str1 + ' ' + string.upper(str2[0]) + str2[1:])


# Prints str to stdout and makes sure the out buffer is flushed.
# This is important for timely feedback over the WEB.
def report(str, ioStream=sys.stdout):
  ioStream.write(str + '\n')
  ioStream.flush()

# Given a string which is the contents of a link field, return an array
# of strings, each with one link info spec. As in:
#  ['title:www:http://db', 'title:frame:/u/paepcke/foo.frm mammals']
# Raises RunnawayArgument

def extractLinks(str):
  res = []
  str = string.strip(str)
  if not str:
    return(res)
  if not (str[0] == '('):
    # If no opening paren, we assume that this is a single link spec comprising
    # the whole field:
    return([str])
  open = 1
  start = 1
  for i in range(len(str)-1):
    char = str[i+1]
    if (open == 0) and (char <> '('):
      # In between parens, just ignore things:
      start = start + 1
      continue
    if char == ')':
      open = open - 1
      if open == 0:
	res.append(string.strip(str[start:i+1]))
	continue
    if char == '(':
      if open == 0:
	start = i + 2
      open = open + 1
      continue
  if open <> 0:
    raise bibimport.RunnawayArgument, "Parens not closed in: " + str
  return(res)

# Given a string of the form 'field:www:http://...', return a list of three
# strings: the field name, the target environment and the URL.
# Raises BadLinkSpecFormat.

def parseLinkSpec(linkSpec):

  if countMatches(linkComponentSeparator, linkSpec) < 2:
    raise BadLinkSpecFormat, "Links should be of the form 'bibField:targetEnv:URL'. Not: '" +\
	  linkSpec + "'"
  res = []
  fieldEnd = string.find(linkSpec, linkComponentSeparator)
  res.append(string.strip(linkSpec[0:fieldEnd]))
  targetSysEnd = string.find(linkSpec, linkComponentSeparator, fieldEnd + 1)
  res.append(string.strip(linkSpec[fieldEnd+1:targetSysEnd]))
  res.append(string.strip(linkSpec[targetSysEnd+1:]))
  # Check that the second spec is really a target environment:
  if (res[1] <> 'www') and (res[1] <> 'frame'):
    raise BadLinkSpecFormat, "Links should be of form 'bibField:targetEnv:URL'. Not: '" + \
	  linkSpec + "'"
  return(res)

# Given any string, return the first substring that is a legal URL, if there is one,
# else return False. We recognize the URL schemes in the first part of
# the urlFinderREPattern in the Constants section at the start of the
# file. Add more there if you wish. 
# Example: "For all good men https://foo.bar can see." returns "https://foo.bar"

def findEmbeddedURL(str):

  matchObj = urlFinder.search(str)
  if matchObj:
    return matchObj.group()
  else:
    return False

# Given one of the three BibTeX string opening chars
# ({, ", or '), return the appropriate closing char.
# If the passed-in char is not a string opener, return
# False:

def getBibTeXStringCloseChar(potentialOpenChar):

  if potentialOpenChar == '"':
    return '"'
  if potentialOpenChar == "'":
    return "'"
  if potentialOpenChar == '{':
    return "}"

  return False
  

# Return number of times a substring occurs in a given string:
def countMatches(subst, str):
  i = 0
  hits = 0
  while 1:
    pos = string.find(str, subst, i)
    if pos == -1:
      return(hits)
    else:
      hits = hits + 1
      i = pos + len(subst)

# Given a single author name string, return it in the form <lastName>, <firstNames>:

def getAuthorLastNameFirst(author):
  author = string.strip(author)
  junior = 0
  jrPos = string.find(author, 'Jr.')
  # Only count this occurrence if the 'Jr.' is at the end of the name.
  # This protects against malfunction for 'Jr. Francis Bell'.
  if (jrPos > -1) and (jrPos == len(author) - 3):
    # Strip the ', Jr.' out (comma may be missing):
    author = string.strip(author[:jrPos])
    if (len(author) > 0) and author[-1] == ',':
      author = author[:-1]
    junior = 1
  commaPos = string.find(author, ",")
  if commaPos > -1:
    # It's of the form 'Miller, Adam', which is fine:
    if junior:
      return(author + ', Jr.')
    else:
      return(author)
  # Find start of last name. It's either after the last period, or after
  # the last space, or at the beginning of the str. Examples:
  # 'R.N. Smith', 'Ronald Smith', 'Smith', 'J. Ray Scott':
  lastNameStartLoc = string.rfind(author, " ")
  if lastNameStartLoc == -1:
    lastNameStartLoc = string.rfind(author, ".")
    if lastNameStartLoc == -1:
      # No comma, period or space in the name: it's just the last name:
      return(author)
  author = string.strip(author[lastNameStartLoc+1:])\
           + ', '\
           + string.strip(author[0:lastNameStartLoc+1])
  if junior:
    author = author + ', Jr.'

  return(author)

# Given a single author name string, return it in the form <firstNames> <lastName>:
def getAuthorLastNameLast(author):
  author = string.strip(author)
  junior = 0
  jrPos = string.find(author, 'Jr.')
  # Only count this occurrence if the 'Jr.' is at the end of the name.
  # This protects against malfunction for 'Jr. Francis Bell'.
  if (jrPos > -1) and (jrPos == len(author) - 3):
    # Strip the ', Jr.' out (comma may be missing):
    author = string.strip(author[:jrPos])
    if (len(author) > 1) and author[-1] == ',':
      author = author[:-1]
    junior = 1
  lastNameEndLoc = string.find(author, ",")
  if lastNameEndLoc > -1:
    # It's of the form 'Miller, Adam'. So it's reversed:
    author = string.strip(author[lastNameEndLoc+1:])\
             + ' '\
             + string.strip(author[:lastNameEndLoc])
  if junior:
    return(author + ', Jr.')
  else:
    return(author)


# Given an entry and a field name, get either the value of the field, if it
# exists, or an empty string:

def getEntryFldSafely(entry, attr):
  try:
    value = getattr(entry, attr)
    return(value)
  except AttributeError:
    return("")
 
# Given an array of bib entries return best guess at alpha order of the
# array by last name:

def sortByLastName(authorEntries):
  return(authorEntries.sort(entryNameComparison))

# Given a dictionary whose values are bib entries, return array with keys in
# order of best guess at alpha order of the entries by last name:

def sortDictBibEntriesByLastName(dict, dictKeys):
  global theBibEntryDict
  theBibEntryDict = dict
  return(dictKeys.sort(dictWrappedEntryNameComparison))
       
def dictWrappedEntryNameComparison(one, other):
  return(entryNameComparison(theBibEntryDict[one], theBibEntryDict[other]))

# Note that this is used as a primitive by other sorting activities as well:

def entryNameComparison(one, other):

  lastNameOne = ''
  lastNameOther = ''

  authorsOne = string.strip(getEntryFldSafely(one, 'author'))
  if not authorsOne:
    authorsOne = string.strip(getEntryFldSafely(one, 'editor'))
    if not authorsOne:
      # If we have neither author nor publisher, alphabetize by title:
      authorsOne = string.strip(getEntryFldSafely(one, 'title'))
      if authorsOne:
        lastNameOne = authorsOne
      else:
	# If we don't even have a title, get it to the front of the list:
	lastNameOne = ' '

      
  authorsOther = string.strip(getEntryFldSafely(other, 'author'))
  if not authorsOther:
    authorsOther = string.strip(getEntryFldSafely(other, 'editor'))
    if not authorsOther:
      authorsOther = string.strip(getEntryFldSafely(other, 'title'))
      if authorsOther:
        lastNameOther = authorsOther
      else:
	lastNameOther = ' '
      
  if not lastNameOne:
    firstAuthorOne = string.splitfields(authorsOne, bibimport.authorSeparator)[0]
    lastNameOne = string.splitfields(getAuthorLastNameFirst(firstAuthorOne), ',')[0]
  if not lastNameOther:
    firstAuthorOther = string.splitfields(authorsOther, bibimport.authorSeparator)[0]
    lastNameOther = string.splitfields(getAuthorLastNameFirst(firstAuthorOther), ',')[0]
  if lastNameOne < lastNameOther:
    return(-1)
  if lastNameOne == lastNameOther:
    return(0)
  else:
    return(1)


# Given an entry and a field name, return a possibly empty list of
# link tags:

def getLinkTags(entry, fldName):

  links = getEntryFldSafely(entry, linksFieldName)

  if not links:

    # No 'links' field in the entry. Check whether
    # instead there is a 'url' field AND caller is asking
    # about the title field. If so, make
    # a WWW link with 'title' as the anchor:

    if fldName == 'title':

      aURL = getEntryFldSafely(entry, urlFieldName)
      if aURL:
	links = [ bibdb.WWWLink(dest=aURL, anchor='title') ]

    else:
      return([])

  res = []

  for link in links:
    if link.anchorField == fldName:
      res.append(link)

  return(res)

# Given an item and a class (not a string, but the class), return 1 if the item
# is an instance of that class:

def isInstanceOf(item, aClass):
  if (type(item) == types.InstanceType) and\
     (item.__class__ == aClass):
    return(1)
  else:
    return(0)
  
# Return an array of the names of all of aClass' subclasses
# (aClass is a class obj, not a str):

def allSubclassesOf(module, aClass):
  subclasses = []
  for name, value in eval(module).__dict__.items():
    if type(value) == types.ClassType:
      if aClass in value.__bases__:
	subclasses.append(name)
  return(subclasses)
										    
def digitsOnly(str):
  """ Is the given string an integer?"""
  try:
     num = int(str)
     return 1
  except ValueError:
     return 0

# Given a file descriptor: read chars till there is EOF or a non-digit. Return
# a string that has all the numbers. FD cursor left pointing at the first non-digit
# byte:

def readNum(fd):
  str = ''
  while 1:
    chr = fd.read(1)
    if chr == '':  # EOF
      return(str)
    if not digitsOnly(chr):
      fd.seek(-1,1)
      return(str)
    str = str + chr

# Given the name of a file that contains a password, return 0 or 1 to reflect
# whether the given password string matches what's in the file. White space
# around both the password in the file, and the password passed in are deleted.
# It is OK for the password file not to exist, if the given password is an
# empty string. If the file exists, but we can't read it (e.g. file permissions),
# we throw a FaultyPasswordMechanism exception:

def verifyPassword(passwd, passwdFilePath):
  if not os.path.isfile(passwdFilePath):
    if (passwd == ""):
      return 1
    else:
      return 0
  try:
    fd = open(passwdFilePath)
    if (string.strip(fd.read()) == passwd):
      fd.close()
      return 1
    else:
      fd.close()
      return 0
  # Could still get error, due to password file permissions:
  except IOError, info:
    raise FaultyPasswordMechanism, "Could not access password information"

# If old passwd is correct, replace the passwd with the new one. Return 1 if
# OK, 0 if permission denied. Throws FaultyPasswordMechanism is can't read/write
# the file containing the password:

def setPassword(dbPath, oldPasswd, newPasswd):
  if verifyPassword(oldPasswd, dbPath):
    try:
      fd = open(dbPath)
      fd.write(newPasswd)
      fd.close()
      return 1
    except IOError, info:
      raise FaultyPasswordMechanism, "Could not access password information"
  else:
    return 0

class KeyGenerator:
  prevKeyTime = 0
  keyUniquifier = 0

  def generateKey(self, prefix='bib'):
    currTime = int(time.time())
    if currTime == self.prevKeyTime:
      self.keyUniquifier = self.keyUniquifier + 1
    else:
      self.keyUniquifier = 0
    key = prefix + str(self.keyUniquifier) + str(currTime)
    self.prevKeyTime = currTime
    return(key)

# Instances of this class can be used to give feedback to the user.  The
# init function can take: a total amount of work to be done (in some
# arbitrary unit), the length of time between reporting to the user (even
# if the 'tick' method is called more often), text to print before the
# amount, text to print after the amount, and whether to report a
# percentage or absolute amount. Example:
#    fg = feedbackGadget(wholeJob = 10)
#    fg.tick(2) => prints '2 done.' if 5 or more seconds have passed since init
# 
#    fg = feedbackGadget(wholeJob = 10,
#		         feedbackLeader = 'Imported ',
#		         feedbackTrailer = ' records.')
#    fg.tick(2) => 'Imported 2 records.'
#
#    fg = feedbackGadget(wholeJob = 10,
#		         feedbackLeader = 'Imported ',
#		         feedbackTrailer = ' of the records.',
#		         reportPercentage = 1)  
#    fg.tick(2) => "Imported 20% of the records."
#
# For progress reporting in consuming files, see subclass fileConsumptionFeedbackGadget

class feedbackGadget:

  def __init__(self,
	       wholeJob = 0,
	       timeTicks = 5.0,
	       feedbackLeader = '',
	       feedbackTrailer = ' done.',
	       reportPercentage = 0):
    self.wholeJob = wholeJob
    self.timeTicks = timeTicks
    self.feedbackLeader = feedbackLeader
    self.feedbackTrailer = feedbackTrailer
    self.reportPercentage = reportPercentage
    self.reset()

  def reset(self):
    self.startTime = time.time()
    self.amountAccomplished = 0

  def tick(self, accomplishment):
    self.amountAccomplished = self.amountAccomplished + accomplishment
    if self.amountAccomplished > self.wholeJob:
      self.amountAccomplished = self.wholeJob
    if (time.time() - self.startTime) > self.timeTicks:
      if self.reportPercentage:
	status = str(int(round((self.amountAccomplished * 100)/self.wholeJob))) + '%'
      else:
	status = str(self.amountAccomplished)
      report(self.feedbackLeader + status + self.feedbackTrailer)
      self.startTime = time.time()

# Use this when you are consuming files. You use it just like the feedbackGadget
# class, except that you pass in a fileFd instead of an absolute wholeJob number,
# and you call tick with no argument.

class fileConsumptionFeedbackGadget(feedbackGadget):
  def __init__(self,
	       fileFd,
	       timeTicks = 5.0,
	       feedbackLeader = '',
	       feedbackTrailer = ' done.',
	       reportPercentage = 0):
    # Check whether it's a StrOrFileFd or a regular fd:
    if type(fileFd.name) <> types.StringType:
      fileName = fileFd.name()
      fileLength = fileFd.len()
    else:
      fileName = fileFd.name
      fileLength = os.stat(fileName)[6]
    feedbackGadget.__init__(self,
			    wholeJob = fileLength,
			    timeTicks = timeTicks,
			    feedbackLeader = feedbackLeader,
			    feedbackTrailer = feedbackTrailer,
			    reportPercentage = reportPercentage)
    self.previousFileCursor = 0
    self.fd = fileFd

  def tick(self):
    newFileCursor = self.fd.tell()
    feedbackGadget.tick(self, newFileCursor - self.previousFileCursor)
    self.previousFileCursor = newFileCursor

class StrOrFileFd:
  pt = 0
  str = ''
  fd  = 0
  fileName = ''
  mode = 'r'
  bufsize = -1
  strOpen = 0
  pastEnd = 0

  def __init__(self, fileName='', str='', mode='r', bufsize=-1):
    self.str = str
    self.fileName = fileName
    self.mode = mode
    self.bufsize = bufsize
    self.pt = 0

    # Will we serve a file?
    if fileName:
      self.fd = open(fileName, mode, bufsize)
    else:
      # Will serve a string:
      self.strOpen = 1
      # make mode be uniformly either r, w, or a (we don't need r+, etc once we've inited things):
      if mode == 'r+':
	self.mode = 'w'
      elif mode == 'r':
	self.mode = mode
      elif (mode == 'w') or (mode == 'w+'):
	self.str = ''
	self.mode = 'w'
      elif (mode == 'a') or (mode == 'a+'):
	# Note that append does leave the str pointer at 0, even though text is appended at end!
	self.mode = 'a'
      else:
	raise ValueError, "Mode must be 'r', 'w', 'a', 'r+', 'w+', or 'a+'"
      self.pastEnd = len(str)
    
  # whether we are representing a file or a string

  def isFile(self):
    if self.fileName:
      return(1)
    else:
      return(0)

  def len(self):
    if self.isFile():
      return(os.stat(self.fileName)[6])
    else:
      return(len(self.str))

  def checkStrClosed(self):
    if not self.strOpen:
      raise ValueError, 'I/O operation on closed file'

  def close(self):
    if self.fileName:
      self.fd.close()
    else:
      self.strOpen = 0
      
  def flush(self):
    if self.fileName:
      self.fd.flush()
    else:
      self.checkStrClosed()

  def isatty(self):
    if self.fileName:
      return(self.fd.isatty())
    else:
      self.checkStrClosed()
      return(0)

  def read(self, size=-1):
    if self.fileName:
      if size < 0:
        return(self.fd.read())
      else:
	return(self.fd.read(size))
    else:
      self.checkStrClosed()
      curPt = self.pt
      if self.pt >= self.pastEnd:
	return('')
      if size < 0:
	self.pt = self.pastEnd
	return(self.str[curPt:])
      else:
	self.pt = self.pt + size
	if self.pt > self.pastEnd:
	  self.pt = self.pastEnd
	  return(self.str[curPt:self.pt])
        else:
          return(self.str[curPt:curPt+size])

  def readline(self):
    if self.fileName:
      return(self.fd.readline())
    else:
      self.checkStrClosed()
      if self.pt >= self.pastEnd:
	return('')
      curPt = self.pt
      endPos = string.find(self.str[self.pt:], '\n')
      if endPos == -1:
	self.pt = self.pastEnd
	return(self.str[curPt:])
      else:
	self.pt = self.pt + endPos + 1
	return(self.str[curPt:curPt + endPos + 1])

  def readlines(self):
    if self.fileName:
      return(self.fd.readlines())
    else:
      self.checkStrClosed()
      if self.pt >= self.pastEnd:
	return('')
      res = []
      while self.pt < self.pastEnd:
	res.append(self.readline())
      return(res)

  def seek(self, offset, whence=0):
    if self.fileName:
      self.fd.seek(offset, whence)
    else:
      self.checkStrClosed()
      if whence == 0:
	self.pt = offset
      elif whence == 1:
	self.pt = self.pt + offset
      elif whence == 2:
	self.pt = (self.pastEnd) + offset
      if self.pt > self.pastEnd:
	self.pt = self.pastEnd
      elif self.pt < 0:
	self.pt = 0

  def tell(self):
    if self.fileName:
      return(self.fd.tell())
    else:
      self.checkStrClosed()
      return(self.pt)

  def write(self, theStr):
    if self.fileName:
      self.fd.write(theStr)
    else:
      if type(theStr) <> types.StringType:
	raise ValueError, "Argument to 'write' must be a string: " + str(theStr)
      self.checkStrClosed()
      if (self.mode <> 'w') and\
	 (self.mode <> 'a'):
	raise IOError, [9, 'Bad file number']
      if self.mode == 'a':
	self.str = self.str + theStr
      else:
	# Need to overwrite at pt:
	self.str = self.str[:self.pt] + theStr + self.str[self.pt + len(theStr):]
      self.pastEnd = len(self.str)
      self.pt = self.pt + len(theStr)
 	
  def writelines(self, list):
    if self.fileName:
      self.fd.writelines(list)
    else:
      self.checkStrClosed()
      for line in list:
	self.write(line)

  def name(self):
    if self.fileName:
      return(self.fd.name)
    else:
      return(self.fileName)
